# -*- coding: utf-8 -*-
"""
    Time Series Forest Regressor (TSF).
"""

__author__ = ["Tony Bagnall", "kkoziara", "luiszugasti", "kanand77", "Markus Löning"]
__all__ = ["TimeSeriesForestRegressor"]

import numpy as np
from joblib import Parallel
from joblib import delayed
from sklearn.ensemble._forest import ForestRegressor
from sklearn.tree import DecisionTreeRegressor

from sktime.regression.base import BaseRegressor
from sktime.series_as_features.base.estimators.interval_based._tsf import (
    BaseTimeSeriesForest,
)
from sktime.series_as_features.base.estimators.interval_based._tsf import _transform
from sktime.utils.validation.panel import check_X


class TimeSeriesForestRegressor(BaseTimeSeriesForest, ForestRegressor, BaseRegressor):
    """Time series forest regressor.

    A time series forest is an ensemble of decision trees built on random intervals.
     Overview: Input n series length m.
     For each tree
         - sample sqrt(m) intervals,
         - find mean, std and slope for each interval, concatenate to form new
         data set,
         - build decision tree on new data set.
     Ensemble the trees with averaged probability estimates.

     This implementation deviates from the original in minor ways. It samples
     intervals with replacement and does not use the splitting criteria tiny
     refinement described in [1]. This is an intentionally stripped down, non
     configurable version for use as a hive-cote component. For a configurable
     tree based ensemble, see sktime.classifiers.ensemble.TimeSeriesForestClassifier

     Parameters
     ----------
     n_estimators    : int, ensemble size, optional (default = 200)
     min_interval    : int, minimum width of an interval, optional (default
     to 3)
     n_jobs          : int, optional (default=1)
         The number of jobs to run in parallel for both `fit` and `predict`.
         ``-1`` means using all processors.
     random_state    : int, seed for random, optional (default = none)

     Attributes
     ----------
     n_classes    : int
     n_intervals  : int
     classes_    : List of classes for a given problem

     References
     ----------
     .. [1] H.Deng, G.Runger, E.Tuv and M.Vladimir, "A time series forest for
     classification and feature extraction",Information Sciences, 239, 2013
     Java implementation
     https://github.com/uea-machine-learning/tsml/blob/master/src/main/
     java/tsml/classifiers/interval_based/TSF.java
     Arxiv version of the paper: https://arxiv.org/abs/1302.2277
    """

    _base_estimator = DecisionTreeRegressor()

    def predict(self, X):
        """Predict

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Panel data

        Returns
        -------
        y_pred : np.array
            Predictions
        """
        self.check_is_fitted()
        X = check_X(X, enforce_univariate=True, coerce_to_numpy=True)
        X = X.squeeze(1)

        _, series_length = X.shape
        if series_length != self.series_length:
            raise TypeError(
                "The number of time points in the training data does not match "
                "that in the test data."
            )
        y_pred = Parallel(n_jobs=self.n_jobs)(
            delayed(_predict)(X, self.estimators_[i], self.intervals_[i])
            for i in range(self.n_estimators)
        )
        return np.mean(y_pred, axis=0)


def _predict(X, estimator, intervals):
    Xt = _transform(X, intervals)
    return estimator.predict(Xt)



# -*- coding: utf-8 -*-
__author__ = ["Markus Löning", "Viktor Kazakov"]
__all__ = ["BaseDataset", "HDDBaseDataset", "BaseResults", "HDDBaseResults"]

import os
from abc import ABC
from abc import abstractmethod
from warnings import warn

import numpy as np
from joblib import dump
from joblib import load


class BaseDataset:
    def __init__(self, name):
        self._name = name

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(name={self.name})"

    def load(self):
        raise NotImplementedError()

    @property
    def name(self):
        return self._name


class HDDBaseDataset(BaseDataset):
    def __init__(self, path, name):
        self._path = path
        super(HDDBaseDataset, self).__init__(name=name)

    @property
    def path(self):
        return self._path

    @staticmethod
    def _validate_path(path):
        """Helper function to validate paths"""
        # check if path already exists
        if not os.path.exists(path):
            raise ValueError(f"No dataset found at path: {path}")


class BaseResults:
    def __init__(self):
        # assigned during fitting of orchestration
        self.strategy_names = []
        self.dataset_names = []
        self.cv = None

    def save_predictions(
        self,
        strategy_name,
        dataset_name,
        y_true,
        y_pred,
        y_proba,
        index,
        cv_fold,
        train_or_test,
    ):
        raise NotImplementedError()

    def load_predictions(self, cv_fold, train_or_test):
        """Loads predictions for all datasets and strategies iteratively"""
        raise NotImplementedError()

    def check_predictions_exist(self, strategy, dataset_name, cv_fold, train_or_test):
        raise NotImplementedError()

    def save_fitted_strategy(self, strategy, dataset_name, cv_fold):
        raise NotImplementedError()

    def load_fitted_strategy(self, strategy_name, dataset_name, cv_fold):
        """Load fitted strategies for all datasets and strategies
        iteratively"""
        raise NotImplementedError()

    def check_fitted_strategy_exists(self, strategy, dataset_name, cv_fold):
        raise NotImplementedError()

    def _append_key(self, strategy_name, dataset_name):
        """Append names of datasets and strategies to results objects during
        orchestration"""
        if strategy_name not in self.strategy_names:
            self.strategy_names.append(strategy_name)

        if dataset_name not in self.dataset_names:
            self.dataset_names.append(dataset_name)

    def _generate_key(self, strategy_name, dataset_name, cv_fold, train_or_test):
        raise NotImplementedError()

    def __repr__(self):
        class_name = self.__class__.__name__
        return (
            f"{class_name}(strategies={self.strategy_names}, datasets="
            f"{self.dataset_names}, "
            f"cv_folds={self.cv.get_n_splits()})"
        )

    def save(self):
        """Save results object as master file"""
        NotImplementedError()

    def _iter(self):
        """Iterate over registry of results object"""
        for strategy_name in self.strategy_names:
            for dataset_name in self.dataset_names:
                yield strategy_name, dataset_name


class HDDBaseResults(BaseResults):
    def __init__(self, path):
        # validate paths
        self._validate_path(path)

        # set path
        self._path = path

        super(HDDBaseResults, self).__init__()

    @property
    def path(self):
        return self._path

    def save(self):
        """Save results object as master file"""
        file = os.path.join(self.path, "results.pickle")

        # if file does not exist already, create a new one
        if not os.path.isfile(file):
            dump(self, file)

        # if file already exists, update file adding new datasets,
        # strategies and/or cv_folds
        else:
            results = load(file)
            self.strategy_names = list(
                set(self.strategy_names + results.strategy_names)
            )
            self.dataset_names = list(set(self.dataset_names + results.dataset_names))
            dump(self, file)

    @staticmethod
    def _validate_path(path):
        """Helper function to validate paths"""
        # check if path already exists
        if os.path.exists(path):
            if not os.path.isdir(path):
                raise ValueError("path already exists and is not a directory")

            elif os.path.isfile(os.path.join(path, "results.pickle")):
                warn(
                    f"Existing results file found in given path: {path}. "
                    f"Results file will be updated"
                )

            elif (
                len([file for file in os.listdir(path) if not file.startswith(".")]) > 0
            ):
                warn("path already exists and is not empty")


class _PredictionsWrapper:
    """Single result class to ensure consistency for return object when
    loading results"""

    def __init__(
        self,
        strategy_name,
        dataset_name,
        index,
        y_true,
        y_pred,
        fit_estimator_start_time,
        fit_estimator_end_time,
        predict_estimator_start_time,
        predict_estimator_end_time,
        y_proba=None,
    ):
        # check input format
        if not all(isinstance(array, np.ndarray) for array in [y_true, y_pred]):
            raise ValueError(
                f"Prediction results have to stored as numpy arrays, "
                f"but found: {[type(array) for array in [y_true, y_pred]]}"
            )
        if not all(isinstance(name, str) for name in [strategy_name, dataset_name]):
            raise ValueError(
                f"Names must be strings, but found: "
                f"{[type(name) for name in [strategy_name, dataset_name]]}"
            )

        self.strategy_name = strategy_name
        self.dataset_name = dataset_name
        self.index = index
        self.y_true = y_true
        self.y_pred = y_pred
        self.y_proba = y_proba
        self.fit_estimator_start_time = fit_estimator_start_time
        self.fit_estimator_end_time = fit_estimator_end_time
        self.predict_estimator_start_time = predict_estimator_start_time
        self.predict_estimator_end_time = predict_estimator_end_time


class BaseMetric(ABC):
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs

    @abstractmethod
    def compute(self, y_true, y_pred):
        """Compute mean and standard error of metric"""



# -*- coding: utf-8 -*-
import numpy as np
from sklearn.model_selection import train_test_split

from sktime.classification.interval_based import SupervisedTimeSeriesForest
from sktime.datasets import load_gunpoint, load_italy_power_demand


def test_y_proba_on_gunpoint():
    X, y = load_gunpoint(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.1, random_state=42
    )
    stsf = SupervisedTimeSeriesForest(random_state=42, n_estimators=20)
    stsf.fit(X_train, y_train)
    actual = stsf.predict_proba(X_test)
    np.testing.assert_array_equal(actual, expected)


def test_stsf_on_power_demand():
    # load power demand data
    X_train, y_train = load_italy_power_demand(split="train", return_X_y=True)
    X_test, y_test = load_italy_power_demand(split="test", return_X_y=True)
    indices = np.random.RandomState(0).permutation(100)

    # train STSF
    stsf = SupervisedTimeSeriesForest(random_state=0, n_estimators=20)
    stsf.fit(X_train, y_train)

    score = stsf.score(X_test.iloc[indices], y_test[indices])
    assert score >= 0.92


# expected y_proba
expected = np.array(
    [
        [
            0.95,
            0.05,
        ],
        [
            1.0,
            0.0,
        ],
        [
            0.95,
            0.05,
        ],
        [
            0.9,
            0.1,
        ],
        [
            0.0,
            1.0,
        ],
        [
            0.9,
            0.1,
        ],
        [
            0.05,
            0.95,
        ],
        [
            1.0,
            0.0,
        ],
        [
            1.0,
            0.0,
        ],
        [
            0.2,
            0.8,
        ],
        [
            0.85,
            0.15,
        ],
        [
            1.0,
            0.0,
        ],
        [
            1.0,
            0.0,
        ],
        [
            0.15,
            0.85,
        ],
        [
            1.0,
            0.0,
        ],
        [
            0.8,
            0.2,
        ],
        [
            1.0,
            0.0,
        ],
        [
            0.95,
            0.05,
        ],
        [
            0.05,
            0.95,
        ],
        [
            0.0,
            1.0,
        ],
    ]
)


# def print_array(array):
#     print('[')
#     for sub_array in array:
#         print('[')
#         for value in sub_array:
#             print(value.astype(str), end='')
#             print(', ')
#         print('],')
#     print(']')
#
#
# if __name__ == "__main__":
#     X, y = load_gunpoint(return_X_y=True)
#     X_train, X_test, y_train, y_test = train_test_split(
#         X, y, test_size=0.1, random_state=42
#     )
#     estimator = SupervisedTimeSeriesForest(random_state=42, n_estimators=20)
#     estimator.fit(X_train, y_train)
#     probas = estimator.predict_proba(X_test)
#     print_array(probas)



# -*- coding: utf-8 -*-
"""Experiments: code to run experiments as an alternative to orchestration.

This file is configured for runs of the main method with command line arguments, or for
single debugging runs. Results are written in a standard format
todo: Tidy up this file!
"""

import os

import sklearn.preprocessing
import sklearn.utils

from sktime.classification.dictionary_based import (
    BOSSEnsemble,
    ContractableBOSS,
    TemporalDictionaryEnsemble,
    WEASEL,
    MUSE,
)
from sktime.classification.distance_based import (
    ElasticEnsemble,
    ProximityForest,
    ProximityTree,
    ProximityStump,
    KNeighborsTimeSeriesClassifier,
    ShapeDTW,
)
from sktime.classification.hybrid import HIVECOTEV1
from sktime.classification.hybrid._catch22_forest_classifier import (
    Catch22ForestClassifier,
)
from sktime.classification.interval_based import (
    TimeSeriesForestClassifier,
    RandomIntervalSpectralForest,
)
from sktime.classification.interval_based._cif import CanonicalIntervalForest
from sktime.classification.interval_based._drcif import DrCIF
from sktime.classification.kernel_based import ROCKETClassifier, Arsenal
from sktime.classification.shapelet_based import MrSEQLClassifier
from sktime.classification.shapelet_based import ShapeletTransformClassifier


os.environ["MKL_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["OMP_NUM_THREADS"] = "1"  # must be done before numpy import!!

import sys
import time
import numpy as np
import pandas as pd


from sklearn import preprocessing
from sklearn.metrics import accuracy_score
from sklearn.model_selection import cross_val_predict
from sktime.utils.data_io import load_from_tsfile_to_dataframe as load_ts
import sktime.datasets.tsc_dataset_names as dataset_lists

__author__ = ["Tony Bagnall"]

"""Prototype mechanism for testing classifiers on the UCR format. This mirrors the
mechanism used in Java,
https://github.com/TonyBagnall/uea-tsc/tree/master/src/main/java/experiments
but is not yet as engineered. However, if you generate results using the method
recommended here, they can be directly and automatically compared to the results
generated in java

"""

classifier_list = [
    # Distance based
    "ProximityForest",
    "KNeighborsTimeSeriesClassifier",
    "ElasticEnsemble",
    "ShapeDTW",
    # Dictionary based
    "BOSS",
    "ContractableBOSS",
    "TemporalDictionaryEnsemble",
    "WEASEL",
    "MUSE",
    # Interval based
    "RandomIntervalSpectralForest",
    "TimeSeriesForestClassifier",
    "CanonicalIntervalForest",
    # Shapelet based
    "ShapeletTransformClassifier",
    "MrSEQLClassifier",
    # Kernel based
    "ROCKET",
    "Arsenal",
]


def set_classifier(cls, resampleId=None):
    """Construct a classifier.

    Basic way of creating the classifier to build using the default settings. This
    set up is to help with batch jobs for multiple problems to facilitate easy
    reproducability. You can set up bespoke classifier in many other ways.

    Parameters
    ----------
    cls: String indicating which classifier you want
    resampleId: classifier random seed

    Return
    ------
    A classifier.
    """
    name = cls.lower()
    # Distance based
    if name == "pf" or name == "proximityforest":
        return ProximityForest(random_state=resampleId)
    elif name == "pt" or name == "proximitytree":
        return ProximityTree(random_state=resampleId)
    elif name == "ps" or name == "proximityStump":
        return ProximityStump(random_state=resampleId)
    elif name == "dtwcv" or name == "kneighborstimeseriesclassifier":
        return KNeighborsTimeSeriesClassifier(distance="dtwcv")
    elif name == "dtw" or name == "1nn-dtw":
        return KNeighborsTimeSeriesClassifier(distance="dtw")
    elif name == "msm" or name == "1nn-msm":
        return KNeighborsTimeSeriesClassifier(distance="msm")
    elif name == "ee" or name == "elasticensemble":
        return ElasticEnsemble()
    elif name == "shapedtw":
        return ShapeDTW()
    # Dictionary based
    elif name == "boss" or name == "bossensemble":
        return BOSSEnsemble(random_state=resampleId)
    elif name == "cboss" or name == "contractableboss":
        return ContractableBOSS(random_state=resampleId)
    elif name == "tde" or name == "temporaldictionaryensemble":
        return TemporalDictionaryEnsemble(random_state=resampleId)
    elif name == "weasel":
        return WEASEL(random_state=resampleId)
    elif name == "muse":
        return MUSE(random_state=resampleId)
    # Interval based
    elif name == "rise" or name == "randomintervalspectralforest":
        return RandomIntervalSpectralForest(random_state=resampleId)
    elif name == "tsf" or name == "timeseriesforestclassifier":
        return TimeSeriesForestClassifier(random_state=resampleId)
    elif name == "cif" or name == "canonicalintervalforest":
        return CanonicalIntervalForest(random_state=resampleId)
    elif name == "drcif":
        return DrCIF(random_state=resampleId)
    # Shapelet based
    elif name == "stc" or name == "shapelettransformclassifier":
        return ShapeletTransformClassifier(
            random_state=resampleId, transform_contract_in_mins=60
        )
    elif name == "mrseql" or name == "mrseqlclassifier":
        return MrSEQLClassifier(seql_mode="fs", symrep=["sax", "sfa"])
    elif name == "rocket":
        return ROCKETClassifier(random_state=resampleId)
    elif name == "arsenal":
        return Arsenal(random_state=resampleId)
    # Hybrid
    elif name == "catch22":
        return Catch22ForestClassifier(random_state=resampleId)
    elif name == "hivecotev1":
        return HIVECOTEV1(random_state=resampleId)
    else:
        raise Exception("UNKNOWN CLASSIFIER")


def stratified_resample(X_train, y_train, X_test, y_test, random_state):
    """Resample data using a random state.

    Reproducable resampling. Combines train and test, resamples to get the same class
    distribution, then returns new trrain and test.

    Parameters
    ----------
    X_train: train data attributes in sktime pandas format.
    y_train: train data class labes as np array.
    X_test: test data attributes in sktime pandas format.
    y_test: test data class labes as np array.

    Returns
    -------
    new train and test attributes and class labels.
    """
    all_labels = np.concatenate((y_train, y_test), axis=None)
    all_data = pd.concat([X_train, X_test])
    random_state = sklearn.utils.check_random_state(random_state)
    # count class occurrences
    unique_train, counts_train = np.unique(y_train, return_counts=True)
    unique_test, counts_test = np.unique(y_test, return_counts=True)
    assert list(unique_train) == list(
        unique_test
    )  # haven't built functionality to deal with classes that exist in
    # test but not in train
    # prepare outputs
    X_train = pd.DataFrame()
    y_train = np.array([])
    X_test = pd.DataFrame()
    y_test = np.array([])
    # for each class
    for label_index in range(0, len(unique_train)):
        # derive how many instances of this class from the counts
        num_instances = counts_train[label_index]
        # get the indices of all instances with this class label
        label = unique_train[label_index]
        indices = np.where(all_labels == label)[0]
        # shuffle them
        random_state.shuffle(indices)
        # take the first lot of instances for train, remainder for test
        train_indices = indices[0:num_instances]
        test_indices = indices[num_instances:]
        del indices  # just to make sure it's not used!
        # extract data from corresponding indices
        train_instances = all_data.iloc[train_indices, :]
        test_instances = all_data.iloc[test_indices, :]
        train_labels = all_labels[train_indices]
        test_labels = all_labels[test_indices]
        # concat onto current data from previous loop iterations
        X_train = pd.concat([X_train, train_instances])
        X_test = pd.concat([X_test, test_instances])
        y_train = np.concatenate([y_train, train_labels], axis=None)
        y_test = np.concatenate([y_test, test_labels], axis=None)
    # get the counts of the new train and test resample
    unique_train_new, counts_train_new = np.unique(y_train, return_counts=True)
    unique_test_new, counts_test_new = np.unique(y_test, return_counts=True)
    # make sure they match the original distribution of data
    assert list(counts_train_new) == list(counts_train)
    assert list(counts_test_new) == list(counts_test)
    return X_train, y_train, X_test, y_test


def run_experiment(
    problem_path,
    results_path,
    cls_name,
    dataset,
    classifier=None,
    resampleID=0,
    overwrite=False,
    format=".ts",
    train_file=False,
):
    """Run a classification experiment.

    Method to run a basic experiment and write the results to files called
    testFold<resampleID>.csv and, if required, trainFold<resampleID>.csv.

    Parameters
    ----------
    problem_path: Location of problem files, full path.
    results_path: Location of where to write results. Any required directories
        will be created
    cls_name: determines which classifier to use, as defined in set_classifier.
        This assumes predict_proba is
    implemented, to avoid predicting twice. May break some classifiers though
    dataset: Name of problem. Files must be  <problem_path>/<dataset>/<dataset>+
                "_TRAIN"+format, same for "_TEST"
    resampleID: Seed for resampling. If set to 0, the default train/test split
                from file is used. Also used in output file name.
    overwrite: if set to False, this will only build results if there is not a
                result file already present. If
    True, it will overwrite anything already there
    format: Valid formats are ".ts", ".arff" and ".long".
    For more info on format, see   examples/Loading%20Data%20Examples.ipynb
    train_file: whether to generate train files or not. If true, it performs a
                10xCV on the train and saves
    """
    build_test = True
    if not overwrite:
        full_path = (
            str(results_path)
            + "/"
            + str(cls_name)
            + "/Predictions/"
            + str(dataset)
            + "/testFold"
            + str(resampleID)
            + ".csv"
        )
        if os.path.exists(full_path):
            print(
                full_path
                + " Already exists and overwrite set to false, not building Test"
            )
            build_test = False
        if train_file:
            full_path = (
                str(results_path)
                + "/"
                + str(cls_name)
                + "/Predictions/"
                + str(dataset)
                + "/trainFold"
                + str(resampleID)
                + ".csv"
            )
            if os.path.exists(full_path):
                print(
                    full_path
                    + " Already exists and overwrite set to false, not building Train"
                )
                train_file = False
        if train_file == False and build_test == False:
            return

    # TO DO: Automatically differentiate between problem types,
    # currently only works with .ts
    trainX, trainY = load_ts(problem_path + dataset + "/" + dataset + "_TRAIN" + format)
    testX, testY = load_ts(problem_path + dataset + "/" + dataset + "_TEST" + format)
    if resampleID != 0:
        # allLabels = np.concatenate((trainY, testY), axis = None)
        # allData = pd.concat([trainX, testX])
        # train_size = len(trainY) / (len(trainY) + len(testY))
        # trainX, testX, trainY, testY = train_test_split(allData, allLabels,
        # train_size=train_size,
        # random_state=resampleID, shuffle=True,
        # stratify=allLabels)
        trainX, trainY, testX, testY = stratified_resample(
            trainX, trainY, testX, testY, resampleID
        )

    le = preprocessing.LabelEncoder()
    le.fit(trainY)
    trainY = le.transform(trainY)
    testY = le.transform(testY)
    if classifier is None:
        classifier = set_classifier(cls_name, resampleID)
    print(cls_name + " on " + dataset + " resample number " + str(resampleID))
    if build_test:
        # TO DO : use sklearn CV
        start = int(round(time.time() * 1000))
        classifier.fit(trainX, trainY)
        build_time = int(round(time.time() * 1000)) - start
        start = int(round(time.time() * 1000))
        probs = classifier.predict_proba(testX)
        preds = classifier.classes_[np.argmax(probs, axis=1)]
        test_time = int(round(time.time() * 1000)) - start
        ac = accuracy_score(testY, preds)
        print(
            cls_name
            + " on "
            + dataset
            + " resample number "
            + str(resampleID)
            + " test acc: "
            + str(ac)
            + " time: "
            + str(test_time)
        )
        #        print(str(classifier.findEnsembleTrainAcc(trainX, trainY)))
        if "Composite" in cls_name:
            second = "Para info too long!"
        else:
            second = str(classifier.get_params())
        second.replace("\n", " ")
        second.replace("\r", " ")

        print(second)
        temp = np.array_repr(classifier.classes_).replace("\n", "")

        third = (
            str(ac)
            + ","
            + str(build_time)
            + ","
            + str(test_time)
            + ",-1,-1,"
            + str(len(classifier.classes_))
        )
        write_results_to_uea_format(
            second_line=second,
            third_line=third,
            output_path=results_path,
            classifier_name=cls_name,
            resample_seed=resampleID,
            predicted_class_vals=preds,
            actual_probas=probs,
            dataset_name=dataset,
            actual_class_vals=testY,
            split="TEST",
        )
    if train_file:
        start = int(round(time.time() * 1000))
        if build_test and hasattr(
            classifier, "_get_train_probs"
        ):  # Normally Can only do this if test has been built
            train_probs = classifier._get_train_probs(trainX)
        else:
            train_probs = cross_val_predict(
                classifier, X=trainX, y=trainY, cv=10, method="predict_proba"
            )
        train_time = int(round(time.time() * 1000)) - start
        train_preds = classifier.classes_[np.argmax(train_probs, axis=1)]
        train_acc = accuracy_score(trainY, train_preds)
        print(
            cls_name
            + " on "
            + dataset
            + " resample number "
            + str(resampleID)
            + " train acc: "
            + str(train_acc)
            + " time: "
            + str(train_time)
        )
        if "Composite" in cls_name:
            second = "Para info too long!"
        else:
            second = str(classifier.get_params())
        second.replace("\n", " ")
        second.replace("\r", " ")
        temp = np.array_repr(classifier.classes_).replace("\n", "")
        third = (
            str(train_acc)
            + ","
            + str(train_time)
            + ",-1,-1,-1,"
            + str(len(classifier.classes_))
        )
        write_results_to_uea_format(
            second_line=second,
            third_line=third,
            output_path=results_path,
            classifier_name=cls_name,
            resample_seed=resampleID,
            predicted_class_vals=train_preds,
            actual_probas=train_probs,
            dataset_name=dataset,
            actual_class_vals=trainY,
            split="TRAIN",
        )


def write_results_to_uea_format(
    output_path,
    classifier_name,
    dataset_name,
    actual_class_vals,
    predicted_class_vals,
    split="TEST",
    resample_seed=0,
    actual_probas=None,
    second_line="No Parameter Info",
    third_line="N/A",
    class_labels=None,
):
    """Write results to file.

    Outputs the classifier results, mirrors that produced by tsml Java package.
    Directories of the form
    <output_path>/<classifier_name>/Predictions/<dataset_name>
    Will automatically be created and results written.

    Parameters
    ----------
    output_path:            string, root path where to put results.
    classifier_name:        string, name of the classifier that made the predictions
    dataset_name:           string, name of the problem the classifier was built on
    actual_class_vals:      array, actual class labels
    predicted_class_vals:   array, predicted class labels
    split:                  string, wither TRAIN or TEST, depending on the results.
    resample_seed:          int, makes resampling deterministic
    actual_probas:          number of cases x number of classes 2d array
    second_line:            unstructured, classifier parameters
    third_line:             summary performance information (see comment below)
    class_labels:           needed to equate to tsml output

    """
    if len(actual_class_vals) != len(predicted_class_vals):
        raise IndexError(
            "The number of predicted class values is not the same as the "
            + "number of actual class values"
        )

    try:
        os.makedirs(
            str(output_path)
            + "/"
            + str(classifier_name)
            + "/Predictions/"
            + str(dataset_name)
            + "/"
        )
    except os.error:
        pass  # raises os.error if path already exists

    if split == "TRAIN" or split == "train":
        train_or_test = "train"
    elif split == "TEST" or split == "test":
        train_or_test = "test"
    else:
        raise ValueError("Unknown 'split' value - should be TRAIN/train or TEST/test")

    file = open(
        str(output_path)
        + "/"
        + str(classifier_name)
        + "/Predictions/"
        + str(dataset_name)
        + "/"
        + str(train_or_test)
        + "Fold"
        + str(resample_seed)
        + ".csv",
        "w",
    )

    # <classifierName>,<datasetName>,<train/test>,<Class Labels>
    file.write(
        str(dataset_name)
        + ","
        + str(classifier_name)
        + ","
        + str(train_or_test)
        + ","
        + str(resample_seed)
        + ",MILLISECONDS,PREDICTIONS, Generated by classification_experiments.py"
    )
    file.write("\n")

    # the second line of the output is free form and classifier-specific;
    # usually this will record info
    # such as parameter options used, any constituent model names for ensembles, etc.
    file.write(str(second_line) + "\n")

    # the third line of the file is the accuracy (should be between 0 and 1 inclusive).
    # If this is a train output file then it will be a training estimate of the
    # classifier on the training data only (e.g. #10-fold cv, leave-one-out cv, etc.).
    # If this is a test output file, it should be the output of the estimator on the
    # test data (likely trained on the training data for a-priori para optimisation)
    file.write(str(third_line))
    file.write("\n")

    # from line 4 onwards each line should include the actual and predicted class labels
    # (comma-separated). If present, for each case, the probabilities of predicting
    # every class value for this case should also be appended to the line (a space is
    # also included between the predicted value and the predict_proba). E.g.:
    # if predict_proba data IS provided for case i:
    #   actual_class_val[i], predicted_class_val[i],,
    #
    # if predict_proba data IS NOT provided for case i:
    #   actual_class_val[i], predicted_class_val[i]
    for i in range(0, len(predicted_class_vals)):
        file.write(str(actual_class_vals[i]) + "," + str(predicted_class_vals[i]))
        if actual_probas is not None:
            file.write(",")
            for j in actual_probas[i]:
                file.write("," + str(j))
            file.write("\n")

    file.close()


def test_loading():
    """Test function to check dataset loading of univariate and multivaria problems."""
    for i in range(0, len(dataset_lists.univariate)):
        data_dir = "E:/tsc_ts/"
        dataset = dataset_lists.univariate[i]
        trainX, trainY = load_ts(data_dir + dataset + "/" + dataset + "_TRAIN.ts")
        testX, testY = load_ts(data_dir + dataset + "/" + dataset + "_TEST.ts")
        print("Loaded " + dataset + " in position " + str(i))
        print("Train X shape :")
        print(trainX.shape)
        print("Train Y shape :")
        print(trainY.shape)
        print("Test X shape :")
        print(testX.shape)
        print("Test Y shape :")
        print(testY.shape)
    for i in range(16, len(dataset_lists.multivariate)):
        data_dir = "E:/mtsc_ts/"
        dataset = dataset_lists.multivariate[i]
        print("Loading " + dataset + " in position " + str(i) + ".......")
        trainX, trainY = load_ts(data_dir + dataset + "/" + dataset + "_TRAIN.ts")
        testX, testY = load_ts(data_dir + dataset + "/" + dataset + "_TEST.ts")
        print("Loaded " + dataset)
        print("Train X shape :")
        print(trainX.shape)
        print("Train Y shape :")
        print(trainY.shape)
        print("Test X shape :")
        print(testX.shape)
        print("Test Y shape :")
        print(testY.shape)


benchmark_datasets = [
    "ACSF1",
    "Adiac",
    "ArrowHead",
    "Beef",
    "BeetleFly",
    "BirdChicken",
    "BME",
    "Car",
    "CBF",
    "ChlorineConcentration",
    "CinCECGTorso",
    "Coffee",
    "Computers",
    "CricketX",
    "CricketY",
    "CricketZ",
    "DiatomSizeReduction",
    "DistalPhalanxOutlineCorrect",
    "DistalPhalanxOutlineAgeGroup",
    "DistalPhalanxTW",
    "Earthquakes",
    "ECG200",
    "ECG5000",
    "ECGFiveDays",
    "EOGHorizontalSignal",
    "EOGVerticalSignal",
    "EthanolLevel",
    "FaceAll",
    "FaceFour",
    "FacesUCR",
    "FiftyWords",
    "Fish",
    "FreezerRegularTrain",
    "FreezerSmallTrain",
    "Ham",
    "Haptics",
    "Herring",
    "InlineSkate",
    "InsectEPGRegularTrain",
    "InsectEPGSmallTrain",
    "InsectWingbeatSound",
    "ItalyPowerDemand",
    "LargeKitchenAppliances",
    "Lightning2",
    "Lightning7",
    "Mallat",
    "Meat",
    "MedicalImages",
    "MiddlePhalanxOutlineCorrect",
    "MiddlePhalanxOutlineAgeGroup",
    "MiddlePhalanxTW",
    "MixedShapesRegularTrain",
    "MixedShapesSmallTrain",
    "MoteStrain",
    "OliveOil",
    "OSULeaf",
    "PhalangesOutlinesCorrect",
    "Phoneme",
    "PigAirwayPressure",
    "PigArtPressure",
    "PigCVP",
    "Plane",
    "PowerCons",
    "ProximalPhalanxOutlineCorrect",
    "ProximalPhalanxOutlineAgeGroup",
    "ProximalPhalanxTW",
    "RefrigerationDevices",
    "Rock",
    "ScreenType",
    "SemgHandGenderCh2",
    "SemgHandMovementCh2",
    "SemgHandSubjectCh2",
    "ShapeletSim",
    "SmallKitchenAppliances",
    "SmoothSubspace",
    "SonyAIBORobotSurface1",
    "SonyAIBORobotSurface2",
    "Strawberry",
    "SwedishLeaf",
    "Symbols",
    "SyntheticControl",
    "ToeSegmentation1",
    "ToeSegmentation2",
    "Trace",
    "TwoLeadECG",
    "TwoPatterns",
    "UMD",
    "UWaveGestureLibraryX",
    "UWaveGestureLibraryY",
    "UWaveGestureLibraryZ",
    "Wafer",
    "Wine",
    "WordSynonyms",
    "Worms",
    "WormsTwoClass",
    "Yoga",
]


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing
    """
    if sys.argv.__len__() > 1:  # cluster run, this is fragile
        print(sys.argv)
        data_dir = sys.argv[1]
        results_dir = sys.argv[2]
        classifier = sys.argv[3]
        dataset = sys.argv[4]
        resample = int(sys.argv[5]) - 1
        tf = str(sys.argv[6]) == "True"
        run_experiment(
            problem_path=data_dir,
            results_path=results_dir,
            cls_name=classifier,
            dataset=dataset,
            resampleID=resample,
            train_file=tf,
        )
    else:  # Local run
        print(" Local Run")
        data_dir = "Z:/ArchiveData/Univariate_ts/"
        results_dir = "Z:/Results Working Area/DistanceBased/sktime/"
        dataset = "ArrowHead"
        trainX, trainY = load_ts(data_dir + dataset + "/" + dataset + "_TRAIN.ts")
        testX, testY = load_ts(data_dir + dataset + "/" + dataset + "_TEST.ts")
        classifier = "1NN-MSM"
        resample = 0
        #         for i in range(0, len(univariate_datasets)):
        #             dataset = univariate_datasets[i]
        # #            print(i)
        # #            print(" problem = "+dataset)
        tf = False
        for i in range(0, len(benchmark_datasets)):
            dataset = benchmark_datasets[i]
            run_experiment(
                overwrite=True,
                problem_path=data_dir,
                results_path=results_dir,
                cls_name=classifier,
                dataset=dataset,
                resampleID=resample,
                train_file=tf,
            )



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from sktime.forecasting.base import BaseForecaster
from sktime.forecasting.base._base import DEFAULT_ALPHA


class Croston(BaseForecaster):
    """Croston's Forecasting Method.

    This was designed for forecasting intermittent demand.

    Parameters
    -----------
    smoothing : float, default = 0.1
        Smoothing parameter

    Examples
    --------
    >>> from sktime.forecasting.croston import Croston
    >>> from sktime.datasets import load_PBS_dataset
    >>> y = load_PBS_dataset()
    >>> forecaster = Croston(smoothing=0.1)
    >>> forecaster.fit(y)
    Croston(...)
    >>> y_pred = forecaster.predict(fh=[1,2,3])

    References
    ----------
    [1]  J. D. Croston. Forecasting and stock control for intermittent demands.
        Operational Research Quarterly (1970-1977), 23(3):pp. 289–303, 1972.
    [2]  Forecasting: Principles and Practice,
        Otext book by Rob J Hyndman and George Athanasopoulos
    """

    _tags = {
        "requires-fh-in-fit": False,  # is forecasting horizon already required in fit?
    }

    def __init__(self, smoothing=0.1):
        # hyperparameter
        self.smoothing = smoothing
        self._f = None
        super(Croston, self).__init__()

    def _fit(self, y, X=None, fh=None):
        """Fit to training data.

        Parameters
        ----------
        y : pd.Series
            Target time series to which to fit the forecaster.
        fh : int, list or np.array, optional (default=None)
            The forecasters horizon with the steps ahead to to predict.
        X : pd.DataFrame, optional (default=None)
            Exogenous variables are ignored
        Returns
        -------
        self : returns an instance of self.
        """
        if X is not None:
            raise NotImplementedError(
                "Support for exogenous variables is not yet implemented"
            )

        n_timepoints = len(y)  # Historical period: i.e the input array's length
        smoothing = self.smoothing

        y = y.to_numpy()  # Transform the input into a numpy array
        # Fit the parameters: level(q), periodicity(a) and forecast(f)
        q, a, f = np.full((3, n_timepoints + 1), np.nan)
        p = 1  # periods since last demand observation

        # Initialization:
        first_occurrence = np.argmax(y[:n_timepoints] > 0)
        q[0] = y[first_occurrence]
        a[0] = 1 + first_occurrence
        f[0] = q[0] / a[0]

        # Create t+1 forecasts:
        for t in range(0, n_timepoints):
            if y[t] > 0:
                q[t + 1] = smoothing * y[t] + (1 - smoothing) * q[t]
                a[t + 1] = smoothing * p + (1 - smoothing) * a[t]
                f[t + 1] = q[t + 1] / a[t + 1]
                p = 1
            else:
                q[t + 1] = q[t]
                a[t + 1] = a[t]
                f[t + 1] = f[t]
                p += 1
        self._f = f

        return self

    def _predict(
        self,
        fh=None,
        X=None,
        return_pred_int=False,
        alpha=DEFAULT_ALPHA,
    ):
        """Predict forecast.

        Parameters
        ----------
        fh : int, list or np.array, optional (default=None)
            The forecasters horizon with the steps ahead to to predict.
        X : pd.DataFrame, optional (default=None)
            Exogenous variables are ignored
        Returns
        -------
        forecast : pd.series
                   predicted forecasts
        """
        if return_pred_int or X is not None:
            raise NotImplementedError()

        len_fh = len(self.fh)
        f = self._f

        # Predicting future forecasts:to_numpy()
        y_pred = np.full(len_fh, f[-1])

        index = self.fh.to_absolute(self.cutoff)
        return pd.Series(y_pred, index=index)



# -*- coding: utf-8 -*-
__all__ = [
    "TimeSeriesForestClassifier",
    "RandomIntervalSpectralForest",
    "SupervisedTimeSeriesForest",
    "CanonicalIntervalForest",
    "DrCIF",
]

from sktime.classification.interval_based._cif import CanonicalIntervalForest
from sktime.classification.interval_based._drcif import DrCIF
from sktime.classification.interval_based._rise import RandomIntervalSpectralForest
from sktime.classification.interval_based._stsf import SupervisedTimeSeriesForest
from sktime.classification.interval_based._tsf import TimeSeriesForestClassifier



# -*- coding: utf-8 -*-
import numpy as np

__author__ = "Jason Lines"


def euclidean_distance(x, y):
    return ed(x, y)


def euclidean(x, y):
    return ed(x, y)


def e_distance(x, y):
    return ed(x, y)


def ed(x, y):
    # convert to numpy array
    x = np.array(x, copy=False)
    y = np.array(y, copy=False)
    # take the sum of squared diffs
    return np.power(x - y, 2).sum()


def dtw_distance(first, second, **kwargs):
    def dtw_single_channel(first, second, **kwargs):
        cutoff = np.inf
        try:
            window = kwargs["window"]
        except Exception:
            window = 1.0
        n = len(first)
        m = len(second)

        warp_matrix = np.full([n, m], np.inf)
        if n > m:
            window_size = n * window
        else:
            window_size = m * window
        window_size = int(window_size)

        def dist(x1, x2):
            return (x1 - x2) ** 2

        pairwise_distances = np.asarray(
            [[dist(x1, x2) for x2 in second] for x1 in first]
        )

        # initialise edges of the warping matrix
        warp_matrix[0][0] = pairwise_distances[0][0]
        for i in range(1, window_size):
            warp_matrix[0][i] = pairwise_distances[0][i] + warp_matrix[0][i - 1]
            warp_matrix[i][0] = pairwise_distances[i][0] + warp_matrix[i - 1][0]

        # now visit all allowed cells, calculate the value as the distance
        # in this cell + min(top, left, or top-left)
        # traverse each row,
        for row in range(1, n):
            cutoff_beaten = False
            # traverse left and right by the allowed window
            for column in range(row - window_size, row + 1 + window_size):
                if column < 1 or column >= m:
                    continue

                # find smallest entry in the warping matrix, either above,
                # to the left, or diagonally left and up
                above = warp_matrix[row - 1][column]
                left = warp_matrix[row][column - 1]
                diag = warp_matrix[row - 1][column - 1]

                # add the pairwise distance for [row][column] to the minimum
                # of the three possible potential cells
                warp_matrix[row][column] = pairwise_distances[row][column] + min(
                    [above, left, diag]
                )

                # check for evidence that cutoff has been beaten on this row
                # (if using)
                if cutoff is not None and warp_matrix[row][column] < cutoff:
                    cutoff_beaten = True

            # if using a cutoff, at least one calculated value on this row
            # MUST be less than the cutoff otherwise the
            # final distance is guaranteed not to be less. Therefore,
            # if the cutoff has not been beaten, early-abandon
            if cutoff is not None and cutoff_beaten is False:
                return float("inf")

        return warp_matrix[n - 1][m - 1]

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return dtw_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += dtw_single_channel(first[dim], second[dim], **kwargs)

    return dist


def derivative_dtw_distance(first, second, **kwargs):
    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return dtw_distance(np.diff(first), np.diff(second), **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += dtw_distance(
            [first[dim].diff()[1:]], [second[dim].diff()[1:]], **kwargs
        )
    return dist


def weighted_dtw_distance(first, second, **kwargs):
    def wdtw_single_channel(first, second, **kwargs):
        try:
            g = kwargs["g"]
        except Exception:
            g = 0.05

        m = len(first)
        n = len(second)

        weight_vector = [1 / (1 + np.exp(-g * (i - m / 2))) for i in range(0, m)]

        def dist(x1, x2):
            return (x1 - x2) ** 2

        pairwise_distances = np.asarray(
            [[dist(x1, x2) for x2 in second] for x1 in first]
        )

        # initialise edges of the warping matrix
        distances = np.full([n, m], np.inf)
        distances[0][0] = weight_vector[0] * pairwise_distances[0][0]

        # top row
        for i in range(1, n):
            distances[0][i] = (
                distances[0][i - 1] + weight_vector[i] * pairwise_distances[0][i]
            )

        # first column
        for i in range(1, m):
            distances[i][0] = (
                distances[i - 1][0] + weight_vector[i] * pairwise_distances[i][0]
            )

        # warp rest
        for i in range(1, m):
            for j in range(1, n):
                min_dist = np.min(
                    [distances[i][j - 1], distances[i - 1][j], distances[i - 1][j - 1]]
                )
                # print(min_dist)
                distances[i][j] = (
                    min_dist + weight_vector[np.abs(i - j)] * pairwise_distances[i][j]
                )
        return distances[m - 1][n - 1]

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return wdtw_single_channel(first, second, **kwargs)
    dist = 0
    for dim in range(0, len(first)):
        dist += wdtw_single_channel(first[dim], second[dim], **kwargs)

    return dist


def weighted_derivative_dtw_distance(first, second, **kwargs):
    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return weighted_dtw_distance(np.diff(first), np.diff(second), **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += weighted_dtw_distance(
            [first[dim].diff()[1:]], [second[dim].diff()[1:]], **kwargs
        )
    return dist


def lcss_distance(first, second, **kwargs):
    def lcss_single_channel(first, second, **kwargs) -> float:
        try:
            delta = kwargs["delta"]
        except KeyError:
            delta = 3

        try:
            epsilon = kwargs["epsilon"]
        except KeyError:
            epsilon = 0.05

        m = len(first)
        n = len(second)

        lcss = np.zeros([m + 1, n + 1])

        for i in range(0, m):
            # for (int j = i-delta j <= i+delta j++){
            for j in range(i - delta, i + delta + 1):
                if j < 0:
                    j = -1
                elif j >= n:
                    j = i + delta
                elif (
                    second[j] + epsilon >= first[i] and second[j] - epsilon <= first[i]
                ):
                    lcss[i + 1][j + 1] = lcss[i][j] + 1
                elif lcss[i][j + 1] > lcss[i + 1][j]:
                    lcss[i + 1][j + 1] = lcss[i][j + 1]
                else:
                    lcss[i + 1][j + 1] = lcss[i + 1][j]

        max_val = -1
        for i in range(1, len(lcss[len(lcss) - 1])):
            if lcss[len(lcss) - 1][i] > max_val:
                max_val = lcss[len(lcss) - 1][i]

        return 1 - (max_val / m)

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return lcss_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += lcss_single_channel(first[dim], second[dim], **kwargs)
    return dist


def msm_distance(first, second, **kwargs):
    def msm_single_channel(first, second, **kwargs) -> float:
        try:
            c = kwargs["c"]
        except KeyError:
            c = 1
        m = len(first)
        n = len(second)
        cost = np.zeros([m, n])

        def calc_cost(new_point, x, y):
            if ((x <= new_point) and (new_point <= y)) or (
                (y <= new_point) and (new_point <= x)
            ):
                return c
            else:
                return c + min(np.abs(new_point - x), np.abs(new_point - y))

        # Initialization
        cost[0][0] = np.abs(first[0] - second[0])
        for i in range(1, m):
            cost[i][0] = cost[i - 1][0] + calc_cost(first[i], first[i - 1], second[0])

        for i in range(1, n):
            cost[0][i] = cost[0][i - 1] + calc_cost(second[i], first[0], second[i - 1])

        # Main Loop
        for i in range(1, m):
            for j in range(1, n):
                d1 = cost[i - 1][j - 1] + np.abs(first[i] - second[j])
                d2 = cost[i - 1][j] + calc_cost(first[i], first[i - 1], second[j])
                d3 = cost[i][j - 1] + calc_cost(second[j], first[i], second[j - 1])
                cost[i][j] = min(d1, d2, d3)

        return cost[m - 1][n - 1]

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return msm_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += msm_single_channel(first[dim], second[dim], **kwargs)
    return dist


def erp_distance(first, second, **kwargs):
    def erp_single_channel(first, second, **kwargs):
        """
        Adapted from:
            This file is part of ELKI:
            Environment for Developing KDD-Applications Supported by
            Index-Structures

            Copyright (C) 2011
            Ludwig-Maximilians-UniversitÃ¤t MÃ¼nchen
            Lehr- und Forschungseinheit fÃ¼r Datenbanksysteme
            ELKI Development Team

            This program is free software: you can redistribute it and/or
            modify
            it under the terms of the GNU Affero General Public License as
            published by
            the Free Software Foundation, either version 3 of the License, or
            (at your option) any later version.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
            GNU Affero General Public License for more details.

            You should have received a copy of the GNU Affero General Public
            License
            along with this program.  If not,
            see <http://www.gnu.org/licenses/>.
        """

        try:
            band_size = kwargs["band_size"]
        except KeyError:
            band_size = 5
        try:
            g = kwargs["g"]
        except KeyError:
            g = 0.5

        m = len(first)
        band = np.ceil(band_size * m)
        curr = np.empty(m)
        prev = np.empty(m)

        for i in range(0, m):
            temp = prev
            prev = curr
            curr = temp
            ll = i - (band + 1)

            if ll < 0:
                ll = 0

            r = i + (band + 1)
            if r > m - 1:
                r = m - 1

            for j in range(ll, r + 1):
                if np.abs(i - j) <= band:
                    val1 = first[i]
                    val2 = g
                    diff = val1 - val2
                    d1 = np.sqrt(diff * diff)

                    val1 = g
                    val2 = second[j]
                    diff = val1 - val2
                    d2 = np.sqrt(diff * diff)

                    val1 = first[i]
                    val2 = second[j]
                    diff = val1 - val2
                    d12 = np.sqrt(diff * diff)

                    dist1 = d1 * d1
                    dist2 = d2 * d2
                    dist12 = d12 * d12

                    if i + j != 0:
                        if i == 0 or (
                            j != 0
                            and (
                                ((prev[j - 1] + dist12) > (curr[j - 1] + dist2))
                                and ((curr[j - 1] + dist2) < (prev[j] + dist1))
                            )
                        ):
                            # del
                            cost = curr[j - 1] + dist2
                        elif (j == 0) or (
                            (i != 0)
                            and (
                                ((prev[j - 1] + dist12) > (prev[j] + dist1))
                                and ((prev[j] + dist1) < (curr[j - 1] + dist2))
                            )
                        ):
                            # ins
                            cost = prev[j] + dist1
                        else:
                            # match
                            cost = prev[j - 1] + dist12
                    else:
                        cost = 0

                    curr[j] = cost

        return np.sqrt(curr[m - 1])

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return erp_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += erp_single_channel(first[dim], second[dim], **kwargs)
    return dist



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-

__author__ = ["Markus Löning"]
__all__ = ["_StatsModelsAdapter"]

import numpy as np
import pandas as pd

from sktime.forecasting.base._base import DEFAULT_ALPHA
from sktime.forecasting.base import BaseForecaster


class _StatsModelsAdapter(BaseForecaster):
    """Base class for interfacing statsmodels forecasting algorithms"""

    _fitted_param_names = ()
    _tags = {
        "univariate-only": True,
        "requires-fh-in-fit": False,
        "handles-missing-data": False,
    }

    def __init__(self):
        self._forecaster = None
        self._fitted_forecaster = None
        super(_StatsModelsAdapter, self).__init__()

    def _fit(self, y, X=None, fh=None):
        """Fit to training data.

        Parameters
        ----------
        y : pd.Series
            Target time series to which to fit the forecaster.
        fh : int, list or np.array, optional (default=None)
            The forecasters horizon with the steps ahead to to predict.
        X : pd.DataFrame, optional (default=None)
            Exogenous variables are ignored
        Returns
        -------
        self : returns an instance of self.
        """
        # statsmodels does not support the pd.Int64Index as required,
        # so we coerce them here to pd.RangeIndex
        if isinstance(y, pd.Series) and type(y.index) == pd.Int64Index:
            y, X = _coerce_int_to_range_index(y, X)
        self._fit_forecaster(y, X)
        return self

    def _fit_forecaster(self, y_train, X_train=None):
        """Internal fit"""
        raise NotImplementedError("abstract method")

    def _predict(self, fh, X=None, return_pred_int=False, alpha=DEFAULT_ALPHA):
        """
        Make forecasts.

        Parameters
        ----------
        fh : ForecastingHorizon
            The forecasters horizon with the steps ahead to to predict.
            Default is one-step ahead forecast,
            i.e. np.array([1])
        X : pd.DataFrame, optional (default=None)
            Exogenous variables are ignored.
        return_pred_int : bool, optional (default=False)
        alpha : int or list, optional (default=0.95)

        Returns
        -------
        y_pred : pd.Series
            Returns series of predicted values.
        """
        if return_pred_int:
            raise NotImplementedError()

        # statsmodels requires zero-based indexing starting at the
        # beginning of the training series when passing integers
        start, end = fh.to_absolute_int(self._y.index[0], self.cutoff)[[0, -1]]
        y_pred = self._fitted_forecaster.predict(start, end)

        # statsmodels forecasts all periods from start to end of forecasting
        # horizon, but only return given time points in forecasting horizon
        return y_pred.loc[fh.to_absolute(self.cutoff).to_pandas()]

    def get_fitted_params(self):
        """Get fitted parameters

        Returns
        -------
        fitted_params : dict
        """
        self.check_is_fitted()
        fitted_params = {}
        for name in self._get_fitted_param_names():
            if name in ["aic", "aicc", "bic", "hqic"]:
                fitted_params[name] = getattr(self._fitted_forecaster, name, None)
            else:
                fitted_params[name] = self._fitted_forecaster.params.get(name)
        return fitted_params

    def _get_fitted_param_names(self):
        """Get names of fitted parameters"""
        return self._fitted_param_names


def _coerce_int_to_range_index(y, X=None):
    new_index = pd.RangeIndex(y.index[0], y.index[-1] + 1)
    try:
        np.testing.assert_array_equal(y.index, new_index)
    except AssertionError:
        raise ValueError(
            "Coercion of pd.Int64Index to pd.RangeIndex "
            "failed. Please provide `y_train` with a "
            "pd.RangeIndex."
        )
    y.index = new_index
    if X is not None:
        X.index = new_index
    return y, X



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = [
    "make_classification_problem",
    "make_regression_problem",
    "make_transformer_problem",
]

import numpy as np
import pandas as pd
from sklearn.utils.validation import check_random_state

from sktime.utils.data_processing import from_3d_numpy_to_nested
from sktime.clustering.tests._clustering_tests import generate_univaritate_series


def _make_panel_X(
    n_instances=20,
    n_columns=1,
    n_timepoints=20,
    y=None,
    return_numpy=False,
    random_state=None,
):
    # If target variable y is given, we ignore n_instances and instead generate as
    # many instances as in the target variable
    if y is not None:
        y = np.asarray(y)
        n_instances = len(y)
    rng = check_random_state(random_state)

    # Generate data as 3d numpy array
    X = rng.normal(scale=0.5, size=(n_instances, n_columns, n_timepoints))

    # Generate association between data and target variable
    if y is not None:
        X = X + (y * 100).reshape(-1, 1, 1)

    if return_numpy:
        return X
    else:
        return from_3d_numpy_to_nested(X)


def _make_regression_y(n_instances=20, return_numpy=True, random_state=None):
    rng = check_random_state(random_state)
    y = rng.normal(size=n_instances)
    if return_numpy:
        return y
    else:
        return pd.Series(y)


def _make_classification_y(
    n_instances=20, n_classes=2, return_numpy=True, random_state=None
):
    if not n_instances > n_classes:
        raise ValueError("n_instances must be bigger than n_classes")
    rng = check_random_state(random_state)
    n_repeats = int(np.ceil(n_instances / n_classes))
    y = np.tile(np.arange(n_classes), n_repeats)[:n_instances]
    rng.shuffle(y)
    if return_numpy:
        return y
    else:
        return pd.Series(y)


def make_classification_problem(
    n_instances=20,
    n_columns=1,
    n_timepoints=20,
    n_classes=2,
    return_numpy=False,
    random_state=None,
):
    y = _make_classification_y(
        n_instances, n_classes, return_numpy=return_numpy, random_state=random_state
    )
    X = _make_panel_X(
        n_columns=n_columns,
        n_timepoints=n_timepoints,
        return_numpy=return_numpy,
        random_state=random_state,
        y=y,
    )

    return X, y


def make_regression_problem(
    n_instances=20, n_columns=1, n_timepoints=20, return_numpy=False, random_state=None
):
    y = _make_regression_y(
        n_instances, random_state=random_state, return_numpy=return_numpy
    )
    X = _make_panel_X(
        n_columns=n_columns,
        n_timepoints=n_timepoints,
        return_numpy=return_numpy,
        random_state=random_state,
        y=y,
    )
    return X, y


def make_clustering_problem(
    n_instances=20, series_size=20, return_numpy=True, random_state=None
):
    # Can only currently support univariate so converting
    # to univaritate for the time being
    X = generate_univaritate_series(n_instances, series_size, random_state)

    if return_numpy:
        return X
    else:
        return pd.Series(X)


def make_transformer_problem(
    n_instances=20,
    n_columns=1,
    n_timepoints=20,
    return_numpy=True,
    random_state=None,
    panel=True,
):
    if not panel:
        X = make_transformer_problem(
            n_instances=n_instances,
            n_columns=n_columns,
            n_timepoints=n_timepoints,
            return_numpy=True,
            random_state=random_state,
            panel=True,
        )
        if return_numpy:
            X = X[0]
        else:
            X = pd.DataFrame(X[0])
    else:
        X = _make_panel_X(
            n_instances=n_instances,
            n_columns=n_columns,
            n_timepoints=n_timepoints,
            return_numpy=True,
            random_state=random_state,
        )
        if not return_numpy:
            arr = []
            for data in X:
                arr.append(pd.DataFrame(data))
            X = arr

    return X


def _make_nested_from_array(array, n_instances=20, n_columns=1):
    return pd.DataFrame(
        [[pd.Series(array) for _ in range(n_columns)] for _ in range(n_instances)],
        columns=[f"col{c}" for c in range(n_columns)],
    )



# -*- coding: utf-8 -*-
"""HIVE-COTE test code."""
import numpy as np

from sktime.classification.hybrid import HIVECOTEV1
from sktime.datasets import load_italy_power_demand


# def test_hivecote_v1_on_gunpoint():
#     # load gunpoint data
#     X_train, y_train = load_gunpoint(split="train", return_X_y=True)
#     X_test, y_test = load_gunpoint(split="test", return_X_y=True)
#     indices = np.random.RandomState(0).permutation(10)
#
#     # train HIVE-COTE v1
#     hc1 = HIVECOTEV1(
#         random_state=0,
#         stc_params={"n_estimators": 10, "transform_contract_in_mins": 0.1},
#         tsf_params={"n_estimators": 10},
#         rise_params={"n_estimators": 10},
#         cboss_params={"n_parameter_samples": 25, "max_ensemble_size": 5},
#     )
#     hc1.fit(X_train.iloc[indices], y_train[indices])
#
#     # assert probabilities are the same
#     probas = hc1.predict_proba(X_test.iloc[indices])
#     testing.assert_array_equal(probas, hivecote_v1_gunpoint_probas)


def test_hivecote_v1_on_power_demand():
    """Test of HIVE-COTEv1 on italy power demand."""
    # load power demand data
    X_train, y_train = load_italy_power_demand(split="train", return_X_y=True)
    X_test, y_test = load_italy_power_demand(split="test", return_X_y=True)
    indices = np.random.RandomState(0).permutation(100)

    # train HIVE-COTE v1
    hc1 = HIVECOTEV1(
        random_state=0,
        stc_params={"n_estimators": 10, "transform_contract_in_mins": 0.1},
        tsf_params={"n_estimators": 10},
        rise_params={"n_estimators": 10},
        cboss_params={"n_parameter_samples": 25, "max_ensemble_size": 5},
    )
    hc1.fit(X_train, y_train)

    score = hc1.score(X_test.iloc[indices], y_test[indices])
    assert score >= 0.92


# hivecote_v1_gunpoint_probas = np.array(
#     []
# )


# def print_array(array):
#     print('[')
#     for sub_array in array:
#         print('[')
#         for value in sub_array:
#             print(value.astype(str), end='')
#             print(', ')
#         print('],')
#     print(']')
#
# if __name__ == "__main__":
#     X_train, y_train = load_gunpoint(split="train", return_X_y=True)
#     X_test, y_test = load_gunpoint(split="test", return_X_y=True)
#     indices = np.random.RandomState(0).permutation(10)
#
# hc1 = HIVECOTEV1(
#     random_state=0,
#     stc_params={"n_estimators": 10, "transform_contract_in_mins": 0.1},
#     tsf_params={"n_estimators": 10},
#     rise_params={"n_estimators": 10},
#     cboss_params={"n_parameter_samples": 25, "max_ensemble_size": 5},
# )
#
#     hc1.fit(X_train.iloc[indices], y_train[indices])
#     probas = hc1.predict_proba(X_test.iloc[indices])
#     print_array(probas)



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from scipy.stats import linregress

from sktime.utils._testing.forecasting import _generate_polynomial_series
from sktime.utils.slope_and_trend import _fit_trend
from sktime.utils.slope_and_trend import _slope


@pytest.mark.parametrize("trend_order", [0, 3])
def test_time_series_slope_against_scipy_linregress(trend_order):
    coefs = np.random.normal(size=(trend_order + 1, 1))
    y = _generate_polynomial_series(20, order=trend_order, coefs=coefs)

    # Compare with scipy's linear regression function
    x = np.arange(y.size) + 1
    a = linregress(x, y).slope
    b = _slope(y)
    np.testing.assert_almost_equal(a, b, decimal=10)


# Check linear and constant cases
@pytest.mark.parametrize("slope", [-1, 0, 1])
def test_time_series_slope_against_simple_cases(slope):
    x = np.arange(1, 10)
    y = x * slope
    np.testing.assert_almost_equal(_slope(y), slope, decimal=10)


@pytest.mark.parametrize("order", [0, 1, 2])  # polynomial order
@pytest.mark.parametrize("n_timepoints", [1, 10])  # number of time series observations
@pytest.mark.parametrize("n_instances", [1, 10])  # number of samples
def test_fit_remove_add_trend(order, n_instances, n_timepoints):
    coefs = np.random.normal(size=order + 1).reshape(-1, 1)
    x = np.column_stack(
        [
            _generate_polynomial_series(n_timepoints, order, coefs=coefs)
            for _ in range(n_instances)
        ]
    ).T
    # assert x.shape == (n_samples, n_obs)

    # check shape of fitted coefficients
    coefs = _fit_trend(x, order=order)
    assert coefs.shape == (n_instances, order + 1)



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-

__author__ = ["Markus Löning"]
__all__ = []

import functools
import warnings


def deprecated(msg=""):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    Parameters
    ----------
    msg : str
        Message for highlighting alternatives or extra info
    """

    def deprecator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn(
                f"{func.__name__} has been deprecated. {msg}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped

    return deprecator



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-

__author__ = ["Markus Löning"]
__all__ = [
    "ShapeletTransformClassifier",
    "MrSEQLClassifier",
    "ROCKETClassifier",
    "BOSSEnsemble",
    "IndividualBOSS",
    "TemporalDictionaryEnsemble",
    "IndividualTDE",
    "KNeighborsTimeSeriesClassifier",
    "ProximityStump",
    "ProximityTree",
    "ProximityForest",
    "ElasticEnsemble",
    "TimeSeriesForestClassifier",
    "RandomIntervalSpectralForest",
    "SupervisedTimeSeriesForest",
    "ComposableTimeSeriesForestClassifier",
    "ColumnEnsembleClassifier",
    "pd",
    "np",
    "load_gunpoint",
    "load_osuleaf",
    "load_basic_motions",
    "load_arrow_head",
]

import numpy as np
import pandas as pd

from sktime.classification.compose import ColumnEnsembleClassifier
from sktime.classification.compose import ComposableTimeSeriesForestClassifier
from sktime.classification.dictionary_based import BOSSEnsemble
from sktime.classification.dictionary_based import IndividualBOSS
from sktime.classification.dictionary_based import IndividualTDE
from sktime.classification.dictionary_based import TemporalDictionaryEnsemble
from sktime.classification.distance_based import ElasticEnsemble
from sktime.classification.distance_based import KNeighborsTimeSeriesClassifier
from sktime.classification.distance_based import ProximityForest
from sktime.classification.distance_based import ProximityStump
from sktime.classification.distance_based import ProximityTree
from sktime.classification.interval_based import RandomIntervalSpectralForest
from sktime.classification.interval_based import SupervisedTimeSeriesForest
from sktime.classification.interval_based import TimeSeriesForestClassifier
from sktime.classification.shapelet_based import MrSEQLClassifier
from sktime.classification.shapelet_based import ShapeletTransformClassifier
from sktime.classification.kernel_based import ROCKETClassifier
from sktime.datasets import load_arrow_head
from sktime.datasets import load_basic_motions
from sktime.datasets import load_gunpoint
from sktime.datasets import load_osuleaf
